@@ -32,6 +32,8 @@
 #include <time.h>
 #include <string.h>
 
+#include <glib/gi18n.h>
+
 #include "server.h"
 
 #include "activation-context.h"
@@ -138,11 +140,8 @@ ac_update_list (ActivationContext *actx, CORBA_Environment *ev)
 			actx->list = NULL;
 		}
 
-		actx->list = Bonobo_ServerInfoList__alloc ();
-		*(actx->list) = cache->_u.server_list;
-		CORBA_sequence_set_release (actx->list, CORBA_FALSE);
-		CORBA_sequence_set_release (&(cache->_u.server_list),
-					    CORBA_FALSE);
+                actx->list = ORBit_copy_value (&cache->_u.server_list,
+                                               TC_Bonobo_ServerInfoList);
 
 		actx->time_list_pulled = time (NULL);
 		actx->by_iid = g_hash_table_new (g_str_hash, g_str_equal);
@@ -164,7 +163,7 @@ ac_query_get_var (Bonobo_ServerInfo *si, const char *id, QueryContext *qctx)
 	retval.value_known = FALSE;
 	retval.needs_free = FALSE;
 
-	if (!strcasecmp (id, "_active")) {
+	if (!g_ascii_strcasecmp (id, "_active")) {
 		CORBA_Environment ev;
 
 		CORBA_exception_init (&ev);
@@ -325,29 +324,55 @@ activation_clients_cache_notify (void)
                 CORBA_Object_release (l2->data, &ev);
                 CORBA_exception_free (&ev);
         }
+
+	g_slist_free (notify);
 }
 
 gboolean
 activation_clients_is_empty_scan (void)
 {
-        GList *l, *next;
+        GList *l, *check = NULL;
 
-        for (l = clients; l; l = next) {
-                next = l->next;
+        for (l = clients; l; l = l->next)
+                check = g_list_prepend (check, CORBA_Object_duplicate (l->data, NULL));
+
+        for (l = check; l; l = l->next) {
                 if (ORBit_small_get_connection_status (l->data) ==
                     ORBIT_CONNECTION_DISCONNECTED) {
-                        CORBA_Object_release (l->data, NULL);
-                        clients = g_list_delete_link (clients, l);
+                        GList *remove;
+                        if ((remove = g_list_find (clients, l->data))) {
+                                CORBA_Object_release (remove->data, NULL);
+                                clients = g_list_delete_link (clients, remove);
+                        }
                 }
+                CORBA_Object_release (l->data, NULL);
         }
 
+        g_list_free (check);
+
         return clients == NULL;
 }
 
-static void
-active_client_cnx_broken (ORBitConnection *cnx,
-                          gpointer         dummy)
+static gboolean
+ac_rescan (gpointer is_idle_rescan)
 {
+        static gboolean in_rescan = FALSE;
+        static guint idle_id = 0;
+
+        server_lock ();
+
+        if (GPOINTER_TO_UINT (is_idle_rescan))
+                idle_id = 0;
+
+        /* We tend to get a lot of 'broken' callbacks at once */
+        if (in_rescan) {
+                if (!idle_id)
+                        idle_id = g_timeout_add (100, ac_rescan, GUINT_TO_POINTER(1));
+                server_unlock ();
+                return FALSE;
+        }
+        in_rescan = TRUE;
+
         if (activation_clients_is_empty_scan ()) {
 #ifdef BONOBO_ACTIVATION_DEBUG
                 g_warning ("All clients dead");
@@ -355,6 +380,17 @@ active_client_cnx_broken (ORBitConnection *cnx,
                 check_quit ();
         }
 
+        in_rescan = FALSE;
+        server_unlock ();
+
+        return FALSE;
+}
+
+static void
+active_client_cnx_broken (ORBitConnection *cnx,
+                          gpointer         dummy)
+{
+        ac_rescan (NULL);
 }
 
 static void
@@ -367,6 +403,8 @@ impl_Bonobo_ActivationContext_addClient (PortableServer_Servant        servant,
         gboolean new_locale;
         ORBitConnection *cnx;
 
+        server_lock ();
+
         new_locale = register_interest_in_locales (locales);
 
         cnx = ORBit_small_get_connection (client);
@@ -378,24 +416,28 @@ impl_Bonobo_ActivationContext_addClient (PortableServer_Servant        servant,
                 clients, CORBA_Object_duplicate (client, ev));
 
         if (!l) {
-                g_signal_connect (
-                        cnx, "broken",
-                        G_CALLBACK (active_client_cnx_broken),
-                        NULL);
+                ORBit_small_listen_for_broken
+                        (client, G_CALLBACK (active_client_cnx_broken), NULL);
                 check_quit ();
         }
 
         if (new_locale)
                 bonobo_object_directory_reload ();
+
+        server_unlock ();
 }
 
 static Bonobo_ObjectDirectoryList *
 impl_Bonobo_ActivationContext__get_directories (PortableServer_Servant servant,
                                                 CORBA_Environment     *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
 	Bonobo_ObjectDirectoryList *retval;
 
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
+
 	retval = Bonobo_ObjectDirectoryList__alloc ();
         if (actx->obj != CORBA_OBJECT_NIL) {
                 retval->_length = 1;
@@ -408,6 +450,8 @@ impl_Bonobo_ActivationContext__get_directories (PortableServer_Servant servant,
 
 	CORBA_sequence_set_release (retval, CORBA_TRUE);
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -416,7 +460,11 @@ impl_Bonobo_ActivationContext_addDirectory (PortableServer_Servant servant,
                                             Bonobo_ObjectDirectory dir,
                                             CORBA_Environment     *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
+
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
 
         if (actx->obj == dir)
                 CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
@@ -426,6 +474,8 @@ impl_Bonobo_ActivationContext_addDirectory (PortableServer_Servant servant,
                 CORBA_Object_release (actx->obj, ev);
                 actx->obj = CORBA_Object_duplicate (dir, ev);
         }
+
+        server_unlock ();
 }
 
 static void
@@ -433,7 +483,11 @@ impl_Bonobo_ActivationContext_removeDirectory (PortableServer_Servant servant,
                                                Bonobo_ObjectDirectory dir,
                                                CORBA_Environment     *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
+
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
 
         if (dir != actx->obj)
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
@@ -446,6 +500,8 @@ impl_Bonobo_ActivationContext_removeDirectory (PortableServer_Servant servant,
                 } else
                         directory_info_free (actx, ev);
         }
+
+        server_unlock ();
 }
 
 static void
@@ -455,11 +511,13 @@ ac_do_activation (ActivationContext                  *actx,
 		  Bonobo_ActivationResult            *out,
 		  Bonobo_ActivationFlags              flags,
 		  const char                         *hostname,
+                  Bonobo_ActivationClient             client,
 		  CORBA_Context                       ctx,
                   CORBA_Environment                  *ev)
 {
 	int num_layers;
 	Bonobo_ServerInfo *activatable;
+        gchar *aid = NULL;
 
 	/* When doing checks for shlib loadability, we 
          * have to find the info on the factory object in case
@@ -502,12 +560,15 @@ ac_do_activation (ActivationContext                  *actx,
 	 * order for loading to work properly (no, we're not going to
 	 * bother with loading a remote shlib into a process - it gets far too complicated
 	 * far too quickly :-) */
+
+        if (activatable)
+                aid = g_strdup_printf ("OAFAID:[%s,%s,%s]", activatable->iid,
+                                       activatable->username, activatable->hostname);
 	
-	if (activatable && !strcmp (activatable->server_type, "shlib")
+	if (activatable && activatable->server_type && !strcmp (activatable->server_type, "shlib")
 	    && !(flags & Bonobo_ACTIVATION_FLAG_NO_LOCAL)
 	    && (hostname && !strcmp (activatable->hostname, hostname))) {
 		int j;
-		char tbuf[512];
 		
 		out->res._d = Bonobo_ACTIVATION_RESULT_SHLIB;		
 
@@ -533,53 +594,53 @@ ac_do_activation (ActivationContext                  *actx,
 
 		/* Copy location into last buffer slot for use in later activation */
 		out->res._u.res_shlib._buffer[j+1] = CORBA_string_dup (activatable->location_info);
-		
-		g_snprintf (tbuf, sizeof (tbuf), "OAFAID:[%s,%s,%s]",
-			    activatable->iid,
-			    activatable->username,
-			    activatable->hostname);
-		out->aid = CORBA_string_dup (tbuf);
+
+		out->aid = CORBA_string_dup (aid);
 	} else {
 		CORBA_Object retval;
+                char *iid = g_strdup (server->iid);
 
 		retval = Bonobo_ObjectDirectory_activate (
-                        actx->obj, server->iid, BONOBO_OBJREF (actx), environment, flags, ctx, ev);
+                        actx->obj, iid, BONOBO_OBJREF (actx),
+                        environment, flags, client, ctx, ev);
+                g_free (iid);
 
 		if (ev->_major == CORBA_NO_EXCEPTION) {
-			char tbuf[512];
 			out->res._d = Bonobo_ACTIVATION_RESULT_OBJECT;
 			out->res._u.res_object = retval;
-			g_snprintf (tbuf, sizeof (tbuf),
-				    "OAFAID:[%s,%s,%s]", activatable->iid,
-				    activatable->username,
-				    activatable->hostname);
-			out->aid = CORBA_string_dup (tbuf);
+			out->aid = CORBA_string_dup (aid);
 		}
 #ifdef BONOBO_ACTIVATION_DEBUG
                 else
                         g_warning ("Activation of '%s' failed with exception '%s'",
-                                   activatable->iid, ev->_id);
+                                   aid, ev->_id);
 #endif
 	}
+        g_free (aid);
 }
 
 
 static Bonobo_ActivationResult *
-impl_Bonobo_ActivationContext_activateMatching (
+impl_Bonobo_ActivationContext_activateMatchingFull (
         PortableServer_Servant              servant,
         const CORBA_char                   *requirements,
         const Bonobo_StringList            *selection_order,
 	const Bonobo_ActivationEnvironment *environment,
         const Bonobo_ActivationFlags        flags,
+        Bonobo_ActivationClient             client,
         CORBA_Context                       ctx,
         CORBA_Environment                  *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
 	Bonobo_ActivationResult *retval = NULL;
 	Bonobo_ServerInfo **items, *curitem;
 	int i;
 	char *hostname;
 
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
+
 	ac_update_lists (actx, ev);
 
 	actx->refs++;
@@ -596,13 +657,15 @@ impl_Bonobo_ActivationContext_activateMatching (
 	retval = Bonobo_ActivationResult__alloc ();
 	retval->res._d = Bonobo_ACTIVATION_RESULT_NONE;
 
+        bonobo_object_ref (actx);
 	for (i = 0; (retval->res._d == Bonobo_ACTIVATION_RESULT_NONE) && items[i]
 	     && (i < actx->total_servers); i++) {
 		curitem = items[i];
 
 		ac_do_activation (actx, curitem, environment,
-				  retval, flags, hostname, ctx, ev);
+				  retval, flags, hostname, client, ctx, ev);
 	}
+        bonobo_object_unref (actx);
 
 	if (retval->res._d == Bonobo_ACTIVATION_RESULT_NONE)
 		retval->aid = CORBA_string_dup ("");
@@ -617,37 +680,24 @@ impl_Bonobo_ActivationContext_activateMatching (
 
 	actx->refs--;
 
+        server_unlock ();
+
 	return retval;
 }
 
-
-static Bonobo_ServerInfoList *
-impl_Bonobo_ActivationContext__get_servers (PortableServer_Servant servant,
-                                            CORBA_Environment     *ev)
+static Bonobo_ActivationResult *
+impl_Bonobo_ActivationContext_activateMatching (
+        PortableServer_Servant              servant,
+        const CORBA_char                   *requirements,
+        const Bonobo_StringList            *selection_order,
+	const Bonobo_ActivationEnvironment *environment,
+        const Bonobo_ActivationFlags        flags,
+        CORBA_Context                       ctx,
+        CORBA_Environment                  *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
-	Bonobo_ServerInfoList *retval;
-	int i;
-	int total;
-
-	ac_update_lists (actx, ev);
-
-	total = actx->total_servers;
-
-	retval = Bonobo_ServerInfoList__alloc ();
-	retval->_length = total;
-	retval->_buffer = CORBA_sequence_Bonobo_ServerInfo_allocbuf (total);
-	CORBA_sequence_set_release (retval, CORBA_TRUE);
-
-	for (i = 0; i < total;) {
-                int j;
-
-                for (j = 0; j < actx->list->_length; j++, i++)
-                        Bonobo_ServerInfo_copy (&retval->_buffer[i],
-                                                &actx->list->_buffer[j]);
-	}
-
-	return retval;
+        return impl_Bonobo_ActivationContext_activateMatchingFull
+                (servant, requirements, selection_order,
+                 environment, flags, CORBA_OBJECT_NIL, ctx, ev);
 }
 
 static Bonobo_ServerInfoList *
@@ -656,12 +706,16 @@ impl_Bonobo_ActivationContext_query (PortableServer_Servant servant,
                                      const Bonobo_StringList * selection_order,
                                      CORBA_Context ctx, CORBA_Environment * ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
 	Bonobo_ServerInfoList *retval;
 	Bonobo_ServerInfo **items;
 	int item_count;
 	int i, j, total;
 
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
+
 	retval = Bonobo_ServerInfoList__alloc ();
 	retval->_length = 0;
 	retval->_buffer = NULL;
@@ -699,6 +753,8 @@ impl_Bonobo_ActivationContext_query (PortableServer_Servant servant,
 
 	actx->refs--;
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -749,9 +805,12 @@ ac_context_to_string_array (CORBA_Context context, char **sort_criteria,
 
         context_username = ac_CORBA_Context_get_value (context, "username", ev);
         context_hostname = ac_CORBA_Context_get_value (context, "hostname", ev);
-	if (ev->_major != CORBA_NO_EXCEPTION)
+	if (ev->_major != CORBA_NO_EXCEPTION) {
+                g_free (context_username);
+                g_free (context_hostname);
                 return;
-        
+        }
+
         sort_criteria[0] = g_strconcat ("username == \'", context_username, "\'", NULL);
         sort_criteria[1] = g_strconcat ("hostname == \'", context_hostname, "\'", NULL);
         sort_criteria[2] = NULL;
@@ -763,19 +822,24 @@ ac_context_to_string_array (CORBA_Context context, char **sort_criteria,
 #define PARSE_ERROR_NOT_AN_AID (_("Not a valid Activation ID"))
 
 static Bonobo_ActivationResult *
-impl_Bonobo_ActivationContext_activateFromAid (PortableServer_Servant  servant,
-					       const CORBA_char       *aid,
-					       Bonobo_ActivationFlags  flags,
-					       CORBA_Context           ctx,
-					       CORBA_Environment      *ev)
+impl_Bonobo_ActivationContext_activateFromAidFull (PortableServer_Servant  servant,
+                                                   const CORBA_char       *aid,
+                                                   Bonobo_ActivationFlags  flags,
+                                                   Bonobo_ActivationClient client,
+                                                   CORBA_Context           ctx,
+                                                   CORBA_Environment      *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
-	Bonobo_ActivationResult *retval;
+        ActivationContext *actx;
+	Bonobo_ActivationResult *retval = NULL;
         char *requirements;
         char *sort_criteria[3];
         Bonobo_StringList selection_order;
 	Bonobo_ActivationEnvironment environment;
 
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
+
 	if (strncmp ("OAFAID:", aid, 7) != 0) {
 		Bonobo_Activation_ParseFailed *ex;
 
@@ -785,26 +849,26 @@ impl_Bonobo_ActivationContext_activateFromAid (PortableServer_Servant  servant,
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
 				     ex_Bonobo_Activation_ParseFailed,
 				     ex);
-		return NULL;
+                goto act_out;
 	}
 
 	ac_update_lists (actx, ev);
         if (ev->_major != CORBA_NO_EXCEPTION)
-                return NULL;
+                goto act_out;
 
 	actx->refs++;
 
         requirements = ac_aid_to_query_string (aid);
         if (requirements == NULL) {
                 actx->refs--;
-                return NULL;
+                goto act_out;
         }
 
         ac_context_to_string_array (ctx, sort_criteria, ev);
         if (ev->_major != CORBA_NO_EXCEPTION) {
                 actx->refs--;
                 g_free (requirements);
-                return NULL;
+                goto act_out;
         }
 
         selection_order._length = 2;
@@ -813,8 +877,9 @@ impl_Bonobo_ActivationContext_activateFromAid (PortableServer_Servant  servant,
 
 	memset (&environment, 0, sizeof (Bonobo_ActivationEnvironment));
 
-        retval = impl_Bonobo_ActivationContext_activateMatching (
-                actx, requirements, &selection_order, &environment, flags, ctx, ev);
+        retval = impl_Bonobo_ActivationContext_activateMatchingFull (
+                actx, requirements, &selection_order, &environment,
+                flags, client, ctx, ev);
 
         g_free (sort_criteria[0]);
         g_free (sort_criteria[1]);
@@ -822,9 +887,32 @@ impl_Bonobo_ActivationContext_activateFromAid (PortableServer_Servant  servant,
 
         actx->refs--;
 
+ act_out:
+        server_unlock ();
+
         return retval;
 }
 
+static Bonobo_ActivationResult *
+impl_Bonobo_ActivationContext_activateFromAid (PortableServer_Servant  servant,
+					       const CORBA_char       *aid,
+					       Bonobo_ActivationFlags  flags,
+					       CORBA_Context           ctx,
+					       CORBA_Environment      *ev)
+{
+        return impl_Bonobo_ActivationContext_activateFromAidFull
+                (servant, aid, flags, CORBA_OBJECT_NIL, ctx, ev);
+}
+
+static CORBA_long
+impl_Bonobo_ActivationContext_getVersion (PortableServer_Servant  servant,
+                                          CORBA_Environment      *ev)
+{
+        return (BONOBO_ACTIVATION_MAJOR_VERSION*10000 + 
+                BONOBO_ACTIVATION_MINOR_VERSION*100 +
+                BONOBO_ACTIVATION_MICRO_VERSION);
+}
+
 static ActivationContext *main_ac = NULL;
 
 void
@@ -832,7 +920,8 @@ activation_context_setup (PortableServer_POA     poa,
                           Bonobo_ObjectDirectory dir,
                           CORBA_Environment     *ev)
 {
-        main_ac = g_object_new (activation_context_get_type (), NULL);
+        main_ac = g_object_new (activation_context_get_type (),
+                                "poa", poa, NULL);
 
         impl_Bonobo_ActivationContext_addDirectory
                 (BONOBO_OBJECT (main_ac), dir, ev);
@@ -882,13 +971,15 @@ activation_context_class_init (ActivationContextClass *klass)
         object_class->finalize = activation_context_finalize;
 
         epv->_get_directories = impl_Bonobo_ActivationContext__get_directories;
-	epv->_get_servers     = impl_Bonobo_ActivationContext__get_servers;
 	epv->addClient         = impl_Bonobo_ActivationContext_addClient;
 	epv->addDirectory      = impl_Bonobo_ActivationContext_addDirectory;
 	epv->removeDirectory   = impl_Bonobo_ActivationContext_removeDirectory;
 	epv->query             = impl_Bonobo_ActivationContext_query;
 	epv->activateMatching  = impl_Bonobo_ActivationContext_activateMatching;
 	epv->activateFromAid   = impl_Bonobo_ActivationContext_activateFromAid;
+	epv->getVersion        = impl_Bonobo_ActivationContext_getVersion;
+	epv->activateMatchingFull = impl_Bonobo_ActivationContext_activateMatchingFull;
+	epv->activateFromAidFull  = impl_Bonobo_ActivationContext_activateFromAidFull;
 }
 
 static void
