@@ -54,6 +54,20 @@
 
 #include <glib/gstdio.h>
 
+#ifdef HAVE_DBUS
+#include <dbus/dbus-glib-lowlevel.h>
+#endif
+
+#ifdef G_OS_WIN32
+#include <io.h>
+#include <conio.h>
+#define _WIN32_WINNT 0x0500 
+#include <windows.h>
+
+static gboolean allocated_new_console = FALSE;
+
+#endif
+
 static gboolean        server_threaded = FALSE;
 static glong           server_guard_depth = 0;
 static GStaticRecMutex server_guard = G_STATIC_REC_MUTEX_INIT;
@@ -183,6 +197,7 @@ static const GOptionEntry options[] = {
         {NULL}
 };
 
+/* Referenced from object-directory-corba.c */
 GMainLoop *main_loop = NULL;
 
 static GString *
@@ -240,6 +255,19 @@ build_src_dir (void)
         return real_od_source_dir;
 }
 
+#ifdef G_OS_WIN32
+
+static void
+wait_console_key (void)
+{
+        SetConsoleTitle ("Bonobo activation server exiting. Type any character to close this window.");
+        printf ("\n"
+                "(Bonobo activation server exiting. Type any character to close this window)\n");
+        _getch ();
+}
+
+#endif
+
 static void
 log_handler (const gchar *log_domain,
              GLogLevelFlags log_level,
@@ -294,6 +322,10 @@ log_handler (const gchar *log_domain,
 	if (log_level & G_LOG_FLAG_FATAL) {
 		fprintf (stderr, "%s%s", message,
                          (message[strlen (message) - 1] == '\n' ? "" : "\n"));
+#ifdef G_OS_WIN32
+                if (allocated_new_console)
+                        wait_console_key ();
+#endif
 		_exit (1);
 	}
 
@@ -363,9 +395,7 @@ dump_ior (CORBA_ORB orb, int dev_null_fd, CORBA_Environment *ev)
 #endif
         if (server_reg) {
                 char *fname;
-                fname = g_build_filename (ORBit_get_safe_tmp (),
-                                          "bonobo-activation-server-ior",
-                                          NULL);
+                fname = _bonobo_activation_ior_fname_get ();
                 fh = g_fopen (fname, "w+");
 		fprintf (fh, "%s\n", ior);
 		fclose (fh);
@@ -375,6 +405,77 @@ dump_ior (CORBA_ORB orb, int dev_null_fd, CORBA_Environment *ev)
 	CORBA_free (ior);
 }
 
+static void
+cleanup_ior_and_lock_files (void)
+{
+    char *fname;
+
+    fname = _bonobo_activation_ior_fname_get ();
+    g_unlink (fname);
+    g_free (fname);
+
+    fname = _bonobo_activation_lock_fname_get ();
+    g_unlink (fname);
+    g_free (fname);
+}
+
+#ifdef HAVE_DBUS
+
+static DBusHandlerResult
+bus_message_handler (DBusConnection *connection,
+                     DBusMessage    *message,
+                     GMainLoop      *loop)
+{
+        if (dbus_message_is_signal (message,
+                                    DBUS_INTERFACE_LOCAL,
+                                    "Disconnected")) {
+                g_main_loop_quit (loop);
+                return DBUS_HANDLER_RESULT_HANDLED;
+        }
+
+        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}
+
+#endif
+
+#ifdef G_OS_WIN32
+
+static void
+ensure_output_visible (gboolean asked_for_it)
+{
+        if (fileno (stdout) != -1 &&
+            _get_osfhandle (fileno (stdout)) != -1)
+                return;
+
+        typedef BOOL (* WINAPI AttachConsole_t) (DWORD);
+
+        AttachConsole_t p_AttachConsole =
+                (AttachConsole_t) GetProcAddress (GetModuleHandle ("kernel32.dll"), "AttachConsole");
+
+        if (p_AttachConsole != NULL) {
+                if (!AttachConsole (ATTACH_PARENT_PROCESS)) {
+                        if (AllocConsole ())
+                                allocated_new_console = TRUE;
+                }
+
+                freopen ("CONOUT$", "w", stdout);
+                dup2 (fileno (stdout), 1);
+                freopen ("CONOUT$", "w", stderr);
+                dup2 (fileno (stderr), 2);
+
+                if (allocated_new_console) {
+                        SetConsoleTitle ("Bonobo activation server output. You can minimize this window, but don't close it.");
+                        if (asked_for_it)
+                                printf ("You asked for debugging output by setting the BONOBO_ACTIVATION_DEBUG_OUTPUT\n"
+                                        "environment variable, so here it is.\n"
+                                        "\n");
+                        atexit (wait_console_key);
+                }
+        }
+}
+
+#endif
+
 int
 main (int argc, char *argv[])
 {
@@ -398,6 +499,10 @@ main (int argc, char *argv[])
 	const gchar                  *debug_output_env;
 #endif
 	GError                       *error = NULL;
+#ifdef HAVE_DBUS
+        DBusConnection               *connection;
+        DBusError                     bus_error;
+#endif
 
 #ifdef HAVE_SETSID
         /*
@@ -422,6 +527,18 @@ main (int argc, char *argv[])
 #endif
 #endif
 
+        g_thread_init (NULL);
+
+#ifdef BONOBO_ACTIVATION_DEBUG
+	debug_output_env = g_getenv ("BONOBO_ACTIVATION_DEBUG_OUTPUT");
+	if (debug_output_env && debug_output_env[0] != '\0') {
+		output_debug = TRUE;
+#ifdef G_OS_WIN32
+                ensure_output_visible (TRUE);
+#endif
+        }
+#endif
+
 	g_set_prgname ("bonobo-activation-server");
 	ctx = g_option_context_new (NULL);
 	g_option_context_add_main_entries (ctx, options, GETTEXT_PACKAGE);
@@ -435,11 +552,6 @@ main (int argc, char *argv[])
         LIBXML_TEST_VERSION;
 	xmlKeepBlanksDefault(0);
 
-#if 0
-        while (!g_file_test ("/tmp/orbit-go", G_FILE_TEST_EXISTS))
-                sleep (1);
-#endif
-
 #ifdef HAVE_SYSLOG_H
 	syslog_ident = g_strdup_printf ("bonobo-activation-server (%s-%u)", g_get_user_name (), (guint) getpid ());
 
@@ -447,11 +559,6 @@ main (int argc, char *argv[])
 	openlog (syslog_ident, 0, LOG_USER);
 #endif
 
-#ifdef BONOBO_ACTIVATION_DEBUG
-	debug_output_env = g_getenv ("BONOBO_ACTIVATION_DEBUG_OUTPUT");
-	if (debug_output_env && debug_output_env[0] != '\0')
-		output_debug = TRUE;
-#endif
         if (server_reg)
                 output_debug = TRUE;
 
@@ -512,10 +619,14 @@ main (int argc, char *argv[])
 	if (existing != CORBA_OBJECT_NIL)
 		CORBA_Object_release (existing, NULL);
 
-        if (ior_fd < 0 && !server_ac && !server_reg)
-                g_critical ("\n\n-- \nThe bonobo-activation-server must be forked by\n"
-                            "libbonobo-activation, and cannot be run itself.\n"
+        if (ior_fd < 0 && !server_ac && !server_reg) {
+#ifdef G_OS_WIN32
+                ensure_output_visible (FALSE);
+#endif
+                g_critical ("\n\n-- \nThe bonobo-activation-server must be started by\n"
+                            "libbonobo-activation, and cannot be run by itself.\n"
                             "This is due to us doing client side locking.\n-- \n");
+        }
 
         if (server_reg) {
                 g_warning ("Running in user-forked debugging mode");
@@ -535,6 +646,28 @@ main (int argc, char *argv[])
 
 	od_finished_internal_registration (); 
 
+#ifdef HAVE_DBUS
+        dbus_error_init (&bus_error);
+        connection = dbus_bus_get (DBUS_BUS_SESSION, &bus_error);
+
+        if (dbus_error_is_set (&bus_error)) {
+                g_warning ("could not associate with desktop session: %s",
+                           bus_error.message);
+                connection = NULL;
+        } else {
+                GMainContext *main_context;
+
+                main_context = g_main_loop_get_context (main_loop);
+                dbus_connection_setup_with_g_main (connection, main_context);
+
+                if (dbus_connection_add_filter (connection,
+                                                (DBusHandleMessageFunction)
+                                                bus_message_handler, main_loop,
+                                                NULL)) {
+                        dbus_connection_set_exit_on_disconnect (connection, FALSE);
+                }
+        }
+#endif
         if (getenv ("BONOBO_ACTIVATION_DEBUG") == NULL)
                 chdir ("/");
 
@@ -550,6 +683,8 @@ main (int argc, char *argv[])
         bonobo_object_directory_shutdown (threaded_poa, ev);
         activation_context_shutdown ();
 
+        cleanup_ior_and_lock_files ();
+
 #ifdef HAVE_SYSLOG_H
 	closelog ();
 	g_free (syslog_ident);
