@@ -32,9 +32,9 @@
 #include <unistd.h>
 #include <string.h>
 
+#include <glib/gi18n.h>
 #include "server.h"
 #include "object-directory.h"
-#include "bonobo-activation/bonobo-activation-i18n.h"
 #include "bonobo-activation/bonobo-activation-private.h"
 #include "activation-server-corba-extensions.h"
 
@@ -270,8 +270,12 @@ impl_Bonobo_ObjectDirectory__get_servers (
         Bonobo_CacheTime                 only_if_newer,
         CORBA_Environment               *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
-	Bonobo_ServerInfoListCache      *retval;
+	ObjectDirectory *od;
+	Bonobo_ServerInfoListCache *retval;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
         update_registry (od, FALSE);
 
@@ -284,6 +288,8 @@ impl_Bonobo_ObjectDirectory__get_servers (
 					    CORBA_FALSE);
 	}
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -304,8 +310,12 @@ impl_Bonobo_ObjectDirectory_get_active_servers (
         Bonobo_CacheTime                 only_if_newer,
         CORBA_Environment               *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
-	Bonobo_ServerStateCache         *retval;
+	ObjectDirectory *od;
+	Bonobo_ServerStateCache *retval;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
 	retval = Bonobo_ServerStateCache__alloc ();
 
@@ -326,6 +336,8 @@ impl_Bonobo_ObjectDirectory_get_active_servers (
 					    CORBA_TRUE);
 	}
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -352,9 +364,20 @@ od_get_active_server (ObjectDirectory    *od,
 			break;
 		}
         }
-	if (retval != CORBA_OBJECT_NIL &&
-	    !CORBA_Object_non_existent (retval, NULL))
-		return CORBA_Object_duplicate (retval, NULL);
+	if (retval != CORBA_OBJECT_NIL) {
+                gboolean non_existent;
+                ServerLockState state;
+
+                /* With dead objects, this path can cause lengthy
+                   re-connection attempts, blocking all other
+                   activation, so drop the lock. g#512520 */
+                state = server_lock_drop();
+                non_existent = CORBA_Object_non_existent (retval, NULL);
+                server_lock_resume (state);
+
+                if (!non_existent)
+                        return CORBA_Object_duplicate (retval, NULL);
+        }
 
 	return CORBA_OBJECT_NIL;
 }
@@ -377,6 +400,7 @@ od_merge_client_environment (ObjectDirectory                    *od,
         int i, serverinfo_env_idx;
         const Bonobo_ActivationEnvironment *client_env;
         const Bonobo_StringList *serverinfo_env = NULL;
+        ClientContext *client_context;
 
         array = g_array_new (FALSE, FALSE, sizeof (Bonobo_ActivationEnvValue));
 
@@ -384,7 +408,16 @@ od_merge_client_environment (ObjectDirectory                    *od,
         for (i = 0; i < environment->_length; ++i)
                 g_array_append_val (array, environment->_buffer[i]);
 
-        if (client == CORBA_OBJECT_NIL)
+        if (G_UNLIKELY (client == CORBA_OBJECT_NIL))
+                goto exit;
+
+        client_context = ((ClientContext *) g_hash_table_lookup
+                          (od->client_contexts, client));
+        if (G_UNLIKELY (!client_context))
+                goto exit;
+
+        client_env = client_context->env;
+        if (G_UNLIKELY (!client_env))
                 goto exit;
 
         /* scan through server properties */
@@ -405,12 +438,6 @@ od_merge_client_environment (ObjectDirectory                    *od,
                 goto exit;
 
         /* do the actual merging */
-        client_env = ((ClientContext *) g_hash_table_lookup
-                      (od->client_contexts, client))->env;
-
-        if (!client_env)
-                goto exit;
-
         for (serverinfo_env_idx = 0;
              serverinfo_env_idx < serverinfo_env->_length; ++serverinfo_env_idx)
         {
@@ -454,14 +481,16 @@ impl_Bonobo_ObjectDirectory_activate (
 	CORBA_Context                       ctx,
 	CORBA_Environment                  *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
-	CORBA_Object                     retval;
-	Bonobo_ServerInfo               *si;
-	ODActivationInfo                 ai;
-#ifdef BONOBO_ACTIVATION_DEBUG
-	static int                       depth = 0;
-#endif
-        Bonobo_ActivationEnvironment merged_environment;
+	ObjectDirectory   *od;
+	CORBA_Object       retval = NULL;
+	Bonobo_ServerInfo *si;
+	ODActivationInfo   ai;
+
+        Bonobo_ActivationEnvironment merged_environment = { 0, };
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
         od_merge_client_environment (od, (Bonobo_ServerInfo *)
                                      g_hash_table_lookup (od->by_iid, iid),
@@ -474,24 +503,16 @@ impl_Bonobo_ObjectDirectory_activate (
         if (!(flags & Bonobo_ACTIVATION_FLAG_PRIVATE)) {
                 retval = od_get_active_server (od, iid, &merged_environment);
 
-                if (retval != CORBA_OBJECT_NIL) {
-                        g_free (merged_environment._buffer);
-                        return retval;
-                }
+                if (retval != CORBA_OBJECT_NIL)
+                        goto act_out;
         }
 
-	if (flags & Bonobo_ACTIVATION_FLAG_EXISTING_ONLY) {
-		return CORBA_OBJECT_NIL;
-        }
+	if (flags & Bonobo_ACTIVATION_FLAG_EXISTING_ONLY)
+                goto act_out;
 
 #ifdef BONOBO_ACTIVATION_DEBUG
-        {
-                int i;
-                depth++;
-                for (i = 0; i < depth; i++)
-                        fputc (' ', stderr);
-                fprintf (stderr, "Activate '%s'\n", iid);
-        }
+        fprintf (stderr, "thread %p start activate '%s'\n",
+                 g_thread_self(), iid);
 #endif
 
 	ai.ac = ac;
@@ -529,16 +550,17 @@ impl_Bonobo_ObjectDirectory_activate (
         }
 
 #ifdef BONOBO_ACTIVATION_DEBUG
-        {
-                int i;
-                for (i = 0; i < depth; i++)
-                        fputc (' ', stderr);
-                fprintf (stderr, "Activated '%s' = %p\n", iid, retval);
-                depth--;
-        }
+        fprintf (stderr, "thread %p end activate '%s' = %p ['%s']\n",
+                 g_thread_self(), iid, retval,
+                 bonobo_exception_get_text (ev)
+                 );
 #endif
+
+ act_out:
         g_free (merged_environment._buffer);
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -580,7 +602,7 @@ check_quit (void)
 
         if (od->n_active_servers <= RESIDUAL_SERVERS &&
             activation_clients_is_empty_scan ())
-                od->no_servers_timeout = g_timeout_add (
+                od->no_servers_timeout = g_timeout_add_seconds (
                         SERVER_IDLE_QUIT_TIMEOUT, quit_server_timeout, NULL);
 
 	od->time_active_changed = time (NULL);
@@ -664,15 +686,31 @@ prune_dead_servers (gpointer key,
         return !servers->n_servers;
 }
 
-static void
-active_server_cnx_broken (ORBitConnection *cnx,
-                          gpointer         dummy)
-
+static gboolean
+as_rescan (gpointer is_idle_rescan)
 {
-        ObjectDirectory *od = main_dir;
+        ObjectDirectory *od;
+        static gboolean in_rescan = FALSE;
+        static guint idle_id = 0;
 
-        if (!od) /* shutting down */
-                return;
+        server_lock ();
+
+        if (GPOINTER_TO_UINT (is_idle_rescan))
+                idle_id = 0;
+
+        if (!(od = main_dir)) { /* shutting down */
+                server_unlock ();
+                return FALSE;
+        }
+
+        /* We tend to get a lot of 'broken' callbacks at once */
+        if (in_rescan) {
+                if (!idle_id)
+                        idle_id = g_timeout_add (100, as_rescan, GUINT_TO_POINTER (1));
+                server_unlock ();
+                return FALSE;
+        }
+        in_rescan = TRUE;
 
         g_hash_table_foreach_remove (od->active_server_lists,
                                      prune_dead_servers, od);
@@ -682,6 +720,19 @@ active_server_cnx_broken (ORBitConnection *cnx,
 #endif
 
         check_quit ();
+
+        in_rescan = FALSE;
+        server_unlock ();
+
+        return FALSE;
+}
+
+static void
+active_server_cnx_broken (ORBitConnection *cnx,
+                          gpointer         dummy)
+
+{
+        as_rescan (NULL);
 }
 
 static void
@@ -699,10 +750,8 @@ add_active_server (ObjectDirectory    *od,
                         g_object_set_data (
                                 G_OBJECT (cnx), "object_count", GUINT_TO_POINTER (1));
                         
-                        g_signal_connect (
-                                cnx, "broken",
-                                G_CALLBACK (active_server_cnx_broken),
-                                NULL);
+                        ORBit_small_listen_for_broken
+                                (object, G_CALLBACK (active_server_cnx_broken), NULL);
                 }
         } else
                 g_assert (!strcmp (iid, NAMING_CONTEXT_IID) ||
@@ -878,10 +927,15 @@ impl_Bonobo_ObjectDirectory_register_new_full (
         Bonobo_ActivationClient             client,
 	CORBA_Environment                  *ev)
 {
-	ObjectDirectory              *od = OBJECT_DIRECTORY (servant);
+	ObjectDirectory              *od;
 	CORBA_Object                  oldobj;
         Bonobo_ActivationEnvironment  merged_environment;
         Bonobo_ServerInfo const      *serverinfo;
+        Bonobo_RegistrationResult     retval = Bonobo_ACTIVATION_REG_SUCCESS;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
 	oldobj = od_get_active_server (od, iid, environment);
         *existing = oldobj;
@@ -893,15 +947,18 @@ impl_Bonobo_ObjectDirectory_register_new_full (
 	oldobj = od_get_active_server (od, iid, &merged_environment);
 	if (oldobj != CORBA_OBJECT_NIL) {
 		if (!CORBA_Object_non_existent (oldobj, ev)) {
-                        g_free (merged_environment._buffer);
-			return Bonobo_ACTIVATION_REG_ALREADY_ACTIVE;
+                        if (CORBA_Object_is_equivalent (oldobj, obj, ev))
+                                retval = Bonobo_ACTIVATION_REG_SUCCESS;
+                        else
+                                retval = Bonobo_ACTIVATION_REG_ALREADY_ACTIVE;
+                        goto reg_out;
                 }
 	}
 
         if (!serverinfo) {
                 if (!(flags&Bonobo_REGISTRATION_FLAG_NO_SERVERINFO)) {
-                        g_free (merged_environment._buffer);
-                        return Bonobo_ACTIVATION_REG_NOT_LISTED;
+                        retval = Bonobo_ACTIVATION_REG_NOT_LISTED;
+                        goto reg_out;
                 }
         }
 
@@ -910,14 +967,17 @@ impl_Bonobo_ObjectDirectory_register_new_full (
 #endif
 
         add_active_server (od, iid, &merged_environment, obj);
-        g_free (merged_environment._buffer);
 	
-	bonobo_event_source_notify_listeners
+        bonobo_event_source_notify_listeners
                 (od->event_source,
                  "Bonobo/ObjectDirectory:activation:register",
-                 NULL, ev);
+                 NULL, NULL);
 
-	return Bonobo_ACTIVATION_REG_SUCCESS;
+ reg_out:
+        g_free (merged_environment._buffer);
+        server_unlock ();
+
+	return retval;
 }
 
 static Bonobo_RegistrationResult
@@ -943,17 +1003,23 @@ impl_Bonobo_ObjectDirectory_unregister (
 	const CORBA_Object      obj,
 	CORBA_Environment      *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
+	ObjectDirectory *od;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
         if (!remove_active_server (od, iid, obj))
                 CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
                                      ex_Bonobo_ObjectDirectory_NotRegistered,
                                      NULL);
-	else 
+	else
                 bonobo_event_source_notify_listeners
                         (od->event_source,
                          "Bonobo/ObjectDirectory:activation:unregister",
-                         NULL, ev);
+                         NULL, NULL);
+
+        server_unlock ();
 }
 
 static Bonobo_DynamicPathLoadResult 
@@ -962,20 +1028,27 @@ impl_Bonobo_ObjectDirectory_add_path(
 	const CORBA_char *		add_path,
 	CORBA_Environment               *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
+	ObjectDirectory *od;
 	int i, j, dir_num, max;
 	char **add_directoies, **ret;
 	GSList *l, *tmp = NULL;
 	gboolean different;
+        Bonobo_DynamicPathLoadResult retval = Bonobo_DYNAMIC_LOAD_SUCCESS;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
 	if (!od->registry_source_directories) {
 		od->registry_source_directories = split_path_unique (add_path);
-		return Bonobo_DYNAMIC_LOAD_SUCCESS;
+                goto add_path_out;
 	} else
 		add_directoies = split_path_unique (add_path);
 
-	if (!add_directoies)
-		return Bonobo_DYNAMIC_LOAD_ERROR;
+	if (!add_directoies) {
+		retval = Bonobo_DYNAMIC_LOAD_ERROR;
+                goto add_path_out;
+        }
 
 	for (max = i = 0; od->registry_source_directories[i]; i++) {
 		tmp = g_slist_append(tmp,g_strdup(od->registry_source_directories[i]));
@@ -1001,7 +1074,8 @@ impl_Bonobo_ObjectDirectory_add_path(
 	if (max == dir_num) {
 		g_strfreev(add_directoies);
 		g_slist_free(tmp);
-		return Bonobo_DYNAMIC_LOAD_ALREADY_LISTED;
+		retval = Bonobo_DYNAMIC_LOAD_ALREADY_LISTED;
+                goto add_path_out;
 	}
 
 	ret = g_new(char *, max + 1);
@@ -1016,7 +1090,11 @@ impl_Bonobo_ObjectDirectory_add_path(
 
 	od->registry_source_directories = ret;
 	update_registry(od, TRUE);	
-	return Bonobo_DYNAMIC_LOAD_SUCCESS;
+
+ add_path_out:
+        server_unlock ();
+
+	return retval;
 }
 
 static Bonobo_DynamicPathLoadResult 
@@ -1025,15 +1103,22 @@ impl_Bonobo_ObjectDirectory_remove_path(
         const CORBA_char *              remove_path,
         CORBA_Environment               *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
+	ObjectDirectory *od;
 	char **remove_directoies, **ret;
 	int i, j, max;
 	GSList *l, *tmp = NULL;
 	gboolean different;
+        Bonobo_DynamicPathLoadResult retval = Bonobo_DYNAMIC_LOAD_SUCCESS;
+
+        server_lock();
+
+        od = OBJECT_DIRECTORY (servant);
 
 	remove_directoies = split_path_unique (remove_path);
-	if (!remove_directoies)
-		return Bonobo_DYNAMIC_LOAD_ERROR;
+	if (!remove_directoies) {
+                retval = Bonobo_DYNAMIC_LOAD_ERROR;
+		goto rm_path_out;
+        }
 
 	for (max = i = 0; od->registry_source_directories[i]; i++) {
 		different = TRUE;
@@ -1053,7 +1138,8 @@ impl_Bonobo_ObjectDirectory_remove_path(
 	if (max == i) {
 		g_slist_free(tmp);
 		g_strfreev(remove_directoies);
-		return Bonobo_DYNAMIC_LOAD_NOT_LISTED;
+                retval = Bonobo_DYNAMIC_LOAD_NOT_LISTED;
+		goto rm_path_out;
 	}	
 	ret = g_new(char *, max + 1);
 	for (l = tmp, i = 0; l; l = l->next)
@@ -1066,7 +1152,9 @@ impl_Bonobo_ObjectDirectory_remove_path(
 
 	od->registry_source_directories = ret;
 	update_registry(od, TRUE);
-	return Bonobo_DYNAMIC_LOAD_SUCCESS;
+ rm_path_out:
+        server_unlock();
+	return retval;
 }
 
 
@@ -1074,12 +1162,18 @@ static void
 client_cnx_broken (ORBitConnection *cnx,
                    const Bonobo_ActivationClient  client)
 {
-        ObjectDirectory *od = main_dir;
+        ObjectDirectory *od;
         ClientContext *context;
         int i;
-        if (!od) /* shutting down */
+
+        server_lock();
+
+        if (!(od = main_dir)) { /* shutting down */
+                server_unlock();
                 return;
-          /* unregister runtime server definitions */
+        }
+
+        /* unregister runtime server definitions */
         context = g_hash_table_lookup (od->client_contexts, client);
         if (context->runtime_iids) {
                 for (i = 0; i < context->runtime_iids->_length; ++i)
@@ -1127,6 +1221,8 @@ client_cnx_broken (ORBitConnection *cnx,
         g_hash_table_remove (od->client_contexts, client);
         od->time_list_changed = time (NULL);
         activation_clients_cache_notify ();
+
+        server_unlock();
 }
 
 static void
@@ -1137,9 +1233,13 @@ impl_Bonobo_ObjectDirectory_addClientEnv (
         CORBA_Environment             *ev)
 {
         Bonobo_ActivationEnvironment *env;
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
-        int i;
+	ObjectDirectory *od;
         ClientContext *context;
+        int i;
+
+        server_lock();
+
+        od = OBJECT_DIRECTORY (servant);
         
         env = Bonobo_ActivationEnvironment__alloc ();
         env->_length  = env->_maximum = client_env->_length;
@@ -1174,6 +1274,8 @@ impl_Bonobo_ObjectDirectory_addClientEnv (
 
         ORBit_small_listen_for_broken (client, G_CALLBACK (client_cnx_broken),
                                        (gpointer) client);
+
+        server_unlock();
 }
 
 
@@ -1202,7 +1304,8 @@ bonobo_object_directory_init (PortableServer_POA poa,
 {
         g_assert (main_dir == NULL);
 
-        main_dir = g_object_new (OBJECT_TYPE_DIRECTORY, NULL);
+        main_dir = g_object_new (OBJECT_TYPE_DIRECTORY,
+                                 "poa", poa, NULL);
 
         main_dir->registry_source_directories = split_path_unique (registry_path);
         update_registry (main_dir, FALSE);
@@ -1219,36 +1322,16 @@ bonobo_object_directory_shutdown (PortableServer_POA poa,
 CORBA_Object
 bonobo_object_directory_re_check_fn (const Bonobo_ActivationEnvironment *environment,
 				     const char                         *act_iid,
-				     gpointer                            user_data,
-				     CORBA_Environment                  *ev)
+				     gpointer                            user_data)
 {
         CORBA_Object retval;
 
+        server_lock ();
+
         retval = od_get_active_server (
                 main_dir, (Bonobo_ImplementationID) act_iid, environment);
-
-        if (ev->_major != CORBA_NO_EXCEPTION ||
-            retval == CORBA_OBJECT_NIL) {
-                char *msg;
-		Bonobo_GeneralError *errval = Bonobo_GeneralError__alloc ();
-
-                CORBA_exception_free (ev);
-
-                /*
-                 * If this exception blows ( which it will only do with a multi-object )
-                 * factory, you need to ensure you register the object you were activated
-                 * for [use const char *bonobo_activation_iid_get (void); ] is registered
-                 * with bonobo_activation_active_server_register - _after_ any other
-                 * servers are registered.
-                 */
-                msg = g_strdup_printf (_("Race condition activating server '%s'"), act_iid);
-                errval->description = CORBA_string_dup (msg);
-                g_free (msg);
-
-		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
-				     ex_Bonobo_GeneralError, errval);
-                retval = CORBA_OBJECT_NIL;
-        }
+        
+        server_unlock ();
 
         return retval;
 }
